@@ -152,9 +152,15 @@ yelp_docbook_document_dispose (GObject *object)
 {
     YelpDocbookDocumentPrivate *priv = GET_PRIV (object);
 
-    g_object_unref (priv->uri);
+    if (priv->uri) {
+        g_object_unref (priv->uri);
+        priv->uri = NULL;
+    }
 
-    g_object_unref (priv->sections);
+    if (priv->sections) {
+        g_object_unref (priv->sections);
+        priv->sections = NULL;
+    }
 
     G_OBJECT_CLASS (yelp_docbook_document_parent_class)->dispose (object);
 }
@@ -346,6 +352,9 @@ docbook_process (YelpDocbookDocument *docbook)
     priv->xmlcur = xmlDocGetRootElement (xmldoc);
 
     id = xmlGetProp (priv->xmlcur, BAD_CAST "id");
+    if (!id)
+        id = xmlGetNsProp (priv->xmlcur, XML_XML_NAMESPACE, BAD_CAST "id");
+
     if (id) {
         priv->root_id = g_strdup (id);
         yelp_document_set_page_id (document, NULL, (gchar *) id);
@@ -360,7 +369,12 @@ docbook_process (YelpDocbookDocument *docbook)
         yelp_document_set_next_id (document, "//about", "//index");
         /* add the id attribute to the root element with value "index"
          * so when we try to load the document later, it doesn't fail */
-        xmlNewProp (priv->xmlcur, BAD_CAST "id", BAD_CAST "//index");
+        if (priv->xmlcur->ns)
+            xmlNewNsProp (priv->xmlcur,
+                          xmlNewNs (priv->xmlcur, XML_XML_NAMESPACE, BAD_CAST "xml"),
+                          BAD_CAST "id", BAD_CAST "//index");
+        else
+            xmlNewProp (priv->xmlcur, BAD_CAST "id", BAD_CAST "//index");
     }
     yelp_document_set_root_id (document, priv->root_id, priv->root_id);
     yelp_document_set_root_id (document, "//about", priv->root_id);
@@ -454,6 +468,7 @@ docbook_walk (YelpDocbookDocument *docbook)
     xmlNodePtr   cur, old_cur;
     GtkTreeIter  iter;
     GtkTreeIter *old_iter = NULL;
+    gboolean chunkQ;
     YelpDocbookDocumentPrivate *priv = GET_PRIV (docbook);
     YelpDocument *document = YELP_DOCUMENT (docbook);
 
@@ -472,13 +487,23 @@ docbook_walk (YelpDocbookDocument *docbook)
                 }
 
     id = xmlGetProp (priv->xmlcur, BAD_CAST "id");
+    if (!id)
+        id = xmlGetNsProp (priv->xmlcur, XML_XML_NAMESPACE, BAD_CAST "id");
 
     if (docbook_walk_divisionQ (docbook) && !id) {
         /* If id attribute is not present, autogenerate a
          * unique value, and insert it into the in-memory tree */
         g_snprintf (autoidstr, 20, "//autoid-%d", ++autoid);
-        xmlNewProp (priv->xmlcur, BAD_CAST "id", BAD_CAST autoidstr);
-        id = xmlGetProp (priv->xmlcur, BAD_CAST "id"); 
+        if (priv->xmlcur->ns) {
+            xmlNewNsProp (priv->xmlcur,
+                          xmlNewNs (priv->xmlcur, XML_XML_NAMESPACE, BAD_CAST "xml"),
+                          BAD_CAST "id", BAD_CAST autoidstr);
+            id = xmlGetNsProp (priv->xmlcur, XML_XML_NAMESPACE, BAD_CAST "id");
+        }
+        else {
+            xmlNewProp (priv->xmlcur, BAD_CAST "id", BAD_CAST autoidstr);
+            id = xmlGetProp (priv->xmlcur, BAD_CAST "id"); 
+        }
     }
 
     if (docbook_walk_chunkQ (docbook)) {
@@ -524,10 +549,12 @@ docbook_walk (YelpDocbookDocument *docbook)
         yelp_document_set_page_id (document, (gchar *) id, priv->cur_page_id);
     }
 
-    yelp_document_signal (YELP_DOCUMENT (docbook),
-                          priv->cur_page_id,
-                          YELP_DOCUMENT_SIGNAL_INFO,
-                          NULL);
+    chunkQ = docbook_walk_chunkQ (docbook);
+    if (chunkQ)
+        yelp_document_signal (YELP_DOCUMENT (docbook),
+                              priv->cur_page_id,
+                              YELP_DOCUMENT_SIGNAL_INFO,
+                              NULL);
 
     for (cur = priv->xmlcur->children; cur; cur = cur->next) {
         if (cur->type == XML_ELEMENT_NODE) {
@@ -538,7 +565,7 @@ docbook_walk (YelpDocbookDocument *docbook)
     priv->cur_depth--;
     priv->xmlcur = old_cur;
 
-    if (docbook_walk_chunkQ (docbook)) {
+    if (chunkQ) {
         priv->sections_iter = old_iter;
         g_free (priv->cur_page_id);
         priv->cur_page_id = old_page_id;
@@ -675,6 +702,8 @@ docbook_walk_get_title (YelpDocbookDocument *docbook)
             }
             else if (!xmlStrcmp (child->name, BAD_CAST "title"))
                 title_tmp = child;
+            else if (!xmlStrcmp (child->name, BAD_CAST "info"))
+                infos[0] = child;
             else if (!xmlStrcmp (child->name, BAD_CAST infoname))
                 infos[0] = child;
             else if (!xmlStrcmp (child->name, BAD_CAST "blockinfo"))
@@ -766,8 +795,8 @@ transform_finished (YelpTransform       *transform,
        is finalized.   Otherwise, we could crash when YelpTransform frees
        its libxslt resources.
      */
-    g_object_weak_ref (transform,
-                       transform_finalized,
+    g_object_weak_ref ((GObject *) transform,
+                       (GWeakNotify) transform_finalized,
                        docbook);
 
     docuri = yelp_uri_get_document_uri (priv->uri);
@@ -807,10 +836,9 @@ transform_finalized (YelpDocbookDocument *docbook,
 {
     YelpDocbookDocumentPrivate *priv = GET_PRIV (docbook);
  
-   debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_FUNCTION, "entering\n");
 
     if (priv->xmldoc)
 	xmlFreeDoc (priv->xmldoc);
     priv->xmldoc = NULL;
-
 }
