@@ -44,10 +44,13 @@ static gboolean       resolve_final              (YelpUri        *uri);
 static void           resolve_file_uri           (YelpUri        *uri);
 static void           resolve_file_path          (YelpUri        *uri);
 static void           resolve_data_dirs          (YelpUri        *uri,
-                                                  const gchar   **subdirs,
+                                                  const gchar    *subdir,
                                                   const gchar    *docid,
-                                                  const gchar    *pageid);
+                                                  const gchar    *pageid,
+                                                  gboolean        langfirst);
 static void           resolve_ghelp_uri          (YelpUri        *uri);
+static void           resolve_help_uri           (YelpUri        *uri);
+static void           resolve_help_list_uri      (YelpUri        *uri);
 static void           resolve_man_uri            (YelpUri        *uri);
 static void           resolve_info_uri           (YelpUri        *uri);
 static void           resolve_xref_uri           (YelpUri        *uri);
@@ -111,6 +114,9 @@ static const gchar *infosuffix[] = {
     NULL
 };
 
+static const gchar default_info_path[] =
+    "/usr/info:/usr/share/info:/usr/local/info:/usr/local/share/info";
+
 /******************************************************************************/
 
 static void
@@ -258,6 +264,12 @@ resolve_async (YelpUri *uri)
         || g_str_has_prefix (priv->res_arg, "gnome-help:")) {
         resolve_ghelp_uri (uri);
     }
+    else if (g_str_has_prefix (priv->res_arg, "help:")) {
+        resolve_help_uri (uri);
+    }
+    else if (g_str_has_prefix (priv->res_arg, "help-list:")) {
+        resolve_help_list_uri (uri);
+    }
     else if (g_str_has_prefix (priv->res_arg, "file:")) {
         resolve_file_uri (uri);
     }
@@ -290,7 +302,7 @@ resolve_async (YelpUri *uri)
         case YELP_URI_DOCUMENT_TYPE_XHTML:
             resolve_file_path (uri);
             break;
-        case YELP_URI_DOCUMENT_TYPE_TOC:
+        case YELP_URI_DOCUMENT_TYPE_HELP_LIST:
             /* FIXME: what do we do? */
             break;
         case YELP_URI_DOCUMENT_TYPE_SEARCH:
@@ -304,12 +316,18 @@ resolve_async (YelpUri *uri)
     }
     else if (strchr (priv->res_arg, ':')) {
         priv->tmptype = YELP_URI_DOCUMENT_TYPE_EXTERNAL;
-        priv->fulluri = g_strdup (priv->res_arg);
     }
     else {
         resolve_file_path (uri);
     }
 
+    /* We _always_ want to have a non-null fulluri, so check for it
+     * having been set here and, if we can't think of something
+     * better, set it to res_arg. */
+    if (!priv->fulluri) {
+        priv->fulluri = g_strdup (priv->res_arg);
+    }
+
  done:
     g_idle_add ((GSourceFunc) resolve_final, uri);
 }
@@ -363,6 +381,15 @@ yelp_uri_get_document_uri (YelpUri *uri)
     YelpUriPrivate *priv = GET_PRIV (uri);
     if (priv->doctype == YELP_URI_DOCUMENT_TYPE_UNRESOLVED)
         return NULL;
+
+    /* There's some client code where it makes sense to want a
+     * document uri, whether or not it conforms to a scheme we really
+     * understand. For example, we might want to look up whether the
+     * given page is currently being visited. */
+    if ((!priv->docuri) && priv->fulluri) {
+        return g_strdup (priv->fulluri);
+    }
+
     return g_strdup (priv->docuri);
 }
 
@@ -518,9 +545,10 @@ resolve_file_path (YelpUri *uri)
 
 static void
 resolve_data_dirs (YelpUri      *ret,
-                   const gchar **subdirs,
+                   const gchar  *subdir,
                    const gchar  *docid,
-                   const gchar  *pageid)
+                   const gchar  *pageid,
+                   gboolean      langfirst)
 {
     const gchar * const *sdatadirs = g_get_system_data_dirs ();
     const gchar * const *langs = g_get_language_names ();
@@ -542,54 +570,68 @@ resolve_data_dirs (YelpUri      *ret,
     searchpath = g_new0 (gchar *, 10);
 
     for (datadir_i = 0; datadirs[datadir_i]; datadir_i++) {
-        for (subdir_i = 0; subdirs[subdir_i]; subdir_i++) {
-            for (lang_i = 0; langs[lang_i]; lang_i++) {
-                gchar *helpdir = g_strdup_printf ("%s%s%s/%s/%s",
-                                                  datadirs[datadir_i],
-                                                  (datadirs[datadir_i][strlen(datadirs[datadir_i]) - 1] == '/' ? "" : "/"),
-                                                  subdirs[subdir_i],
-                                                  docid,
-                                                  langs[lang_i]);
-                if (!g_file_test (helpdir, G_FILE_TEST_IS_DIR)) {
-                    g_free (helpdir);
-                    continue;
-                }
+        for (lang_i = 0; langs[lang_i]; lang_i++) {
+            gchar *helpdir = g_build_filename (datadirs[datadir_i],
+                                               subdir,
+                                               langfirst ? langs[lang_i] : docid,
+                                               langfirst ? docid : langs[lang_i],
+                                               NULL);
+            if (!g_file_test (helpdir, G_FILE_TEST_IS_DIR)) {
+                g_free (helpdir);
+                continue;
+            }
 
-                if (searchi + 1 >= searchmax) {
-                    searchmax += 5;
-                    searchpath = g_renew (gchar *, searchpath, searchmax);
-                }
-                searchpath[searchi] = helpdir;
-                searchpath[++searchi] = NULL;
+            if (searchi + 1 >= searchmax) {
+                searchmax += 5;
+                searchpath = g_renew (gchar *, searchpath, searchmax);
+            }
+            searchpath[searchi] = helpdir;
+            searchpath[++searchi] = NULL;
 
-                if (priv->tmptype != YELP_URI_DOCUMENT_TYPE_UNRESOLVED)
-                    /* We've already found it.  We're just adding to the search path now. */
-                    continue;
+            if (priv->tmptype != YELP_URI_DOCUMENT_TYPE_UNRESOLVED)
+                /* We've already found it.  We're just adding to the search path now. */
+                continue;
 
-                filename = g_strdup_printf ("%s/index.page", helpdir);
+            filename = g_strdup_printf ("%s/index.page", helpdir);
+            if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
+                priv->tmptype = YELP_URI_DOCUMENT_TYPE_MALLARD;
+                g_free (filename);
+                filename = g_strdup (helpdir);
+                continue;
+            }
+            g_free (filename);
+
+            if (langfirst) {
+                filename = g_strdup_printf ("%s/index.docbook", helpdir);
                 if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
-                    priv->tmptype = YELP_URI_DOCUMENT_TYPE_MALLARD;
-                    g_free (filename);
-                    filename = g_strdup (helpdir);
+                    priv->tmptype = YELP_URI_DOCUMENT_TYPE_DOCBOOK;
                     continue;
                 }
                 g_free (filename);
-
+            }
+            else {
                 filename = g_strdup_printf ("%s/%s.xml", helpdir, pageid);
                 if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
                     priv->tmptype = YELP_URI_DOCUMENT_TYPE_DOCBOOK;
                     continue;
                 }
                 g_free (filename);
+            }
 
-                filename = g_strdup_printf ("%s/%s.html", helpdir, pageid);
-                if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
-                    priv->tmptype = YELP_URI_DOCUMENT_TYPE_HTML;
-                    continue;
-                }
-                g_free (filename);
-            } /* end for langs */
-        } /* end for subdirs */
+            filename = g_strdup_printf ("%s/%s.html", helpdir, pageid);
+            if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
+                priv->tmptype = YELP_URI_DOCUMENT_TYPE_HTML;
+                continue;
+            }
+            g_free (filename);
+
+            filename = g_strdup_printf ("%s/%s.xhtml", helpdir, pageid);
+            if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
+                priv->tmptype = YELP_URI_DOCUMENT_TYPE_XHTML;
+                continue;
+            }
+            g_free (filename);
+        } /* end for langs */
     } /* end for datadirs */
 
     g_free (datadirs);
@@ -610,7 +652,6 @@ resolve_ghelp_uri (YelpUri *uri)
      * ghelp:document[/file][?page][#frag]
      */
     YelpUriPrivate *priv = GET_PRIV (uri);
-    const gchar const *helpdirs[3] = {"help", "gnome/help", NULL};
     gchar *document, *slash, *query, *hash;
     gchar *colon, *c; /* do not free */
 
@@ -659,7 +700,7 @@ resolve_ghelp_uri (YelpUri *uri)
         resolve_file_uri (uri);
     }
     else {
-        resolve_data_dirs (uri, helpdirs, document, slash ? slash : document);
+        resolve_data_dirs (uri, "gnome/help", document, slash ? slash : document, FALSE);
     }
 
     if (query && hash) {
@@ -692,6 +733,84 @@ resolve_ghelp_uri (YelpUri *uri)
     return;
 }
 
+static void
+resolve_help_uri (YelpUri *uri)
+{
+    /* help:document[/page][?query][#frag]
+     */
+    YelpUriPrivate *priv = GET_PRIV (uri);
+    gchar *document, *slash, *query, *hash;
+    gchar *colon, *c; /* do not free */
+
+    colon = strchr (priv->res_arg, ':');
+    if (!colon) {
+        priv->tmptype = YELP_URI_DOCUMENT_TYPE_ERROR;
+        return;
+    }
+
+    slash = query = hash = NULL;
+    for (c = colon; *c != '\0'; c++) {
+        if (*c == '#' && hash == NULL)
+            hash = c;
+        else if (*c == '?' && query == NULL && hash == NULL)
+            query = c;
+        else if (*c == '/' && slash == NULL && query == NULL && hash == NULL)
+            slash = c;
+    }
+
+    if (slash || query || hash)
+        document = g_strndup (colon + 1,
+                              (slash ? slash : (query ? query : hash)) - colon - 1);
+    else
+        document = g_strdup (colon + 1);
+
+    if (slash && (query || hash))
+        slash = g_strndup (slash + 1,
+                           (query ? query : hash) - slash - 1);
+    else if (slash)
+        slash = g_strdup (slash + 1);
+
+    if (query && hash)
+        query = g_strndup (query + 1,
+                           hash - query - 1);
+    else if (query)
+        query = g_strdup (query + 1);
+
+    if (hash)
+        hash = g_strdup (hash + 1);
+
+    priv->page_id = (slash ? slash : g_strdup ("index"));
+    resolve_data_dirs (uri, "help", document, priv->page_id, TRUE);
+
+    if (hash)
+        priv->frag_id = hash;
+
+    priv->docuri = g_strconcat ("help:", document, NULL);
+
+    priv->fulluri = g_strconcat (priv->docuri,
+                                 priv->page_id ? "/" : "",
+                                 priv->page_id ? priv->page_id : "",
+                                 query ? "?" : "",
+                                 query ? query : "",
+                                 priv->frag_id ? "#" : "",
+                                 priv->frag_id ? priv->frag_id : "",
+                                 NULL);
+
+    g_free (query);
+    g_free (document);
+    return;
+}
+
+static void
+resolve_help_list_uri (YelpUri *uri)
+{
+    YelpUriPrivate *priv = GET_PRIV (uri);
+    priv->docuri = g_strdup ("help-list:");
+    priv->fulluri = g_strdup (priv->res_arg);
+    priv->page_id = g_strdup ("index");
+    priv->tmptype = YELP_URI_DOCUMENT_TYPE_HELP_LIST;
+}
+
 static void
 resolve_man_uri (YelpUri *uri)
 {
@@ -842,6 +961,66 @@ resolve_man_uri (YelpUri *uri)
     g_free (newarg);
 }
 
+/*
+  Return 1 if ch is a number from 0 to 9 or a letter a-f or A-F and 0
+  otherwise. This is sort of not utf8-safe, but since we are only
+  looking for 7-bit things, it doesn't matter.
+ */
+static int
+is_hex (gchar ch)
+{
+    if (((48 <= ch) && (ch <= 57)) ||
+        ((65 <= ch) && (ch <= 70)) ||
+        ((97 <= ch) && (ch <= 102)))
+        return 1;
+    return 0;
+}
+
+/*
+  Return a newly allocated string, where %ab for a,b in [0, f] is
+  replaced by the character it represents.
+ */
+static gchar*
+decode_url (const gchar *url)
+{
+    if (!url) return NULL;
+
+    unsigned int len = strlen (url) + 1;
+    int hex;
+    gchar *ret = g_new (gchar, len);
+    const gchar *ptr = url, *end = url+len;
+    gchar *retptr = ret, *tmp;
+
+    while (ptr < end) {
+        if (*ptr == '%' && is_hex(*(ptr + 1)) && is_hex(*(ptr + 2))) {
+            *retptr = *(ptr+1);
+            *(retptr+1) = *(ptr+2);
+            *(retptr+2) = '\0';
+
+            sscanf (retptr, "%x", &hex);
+
+            if (hex < 0 || hex > 127) {
+                g_warning ("Skipping non-7-bit character.");
+                ptr++;
+                continue;
+            }
+            *retptr = (gchar)hex;
+
+            retptr++;
+            ptr += 3;
+        }
+        else {
+            tmp = g_utf8_next_char(ptr);
+            memcpy (retptr, ptr, (tmp-ptr));
+            retptr += tmp-ptr;
+            ptr = tmp;
+        }
+    }
+    *retptr = '\0';
+
+    return ret;
+}
+
 static void
 resolve_info_uri (YelpUri *uri)
 {
@@ -853,7 +1032,6 @@ resolve_info_uri (YelpUri *uri)
      * info:(name)
      */
     static gchar **infopath = NULL;
-    const gchar * const * langs = g_get_language_names ();
     gchar *name = NULL;
     gchar *sect = NULL;
     gchar *fullpath = NULL;
@@ -873,10 +1051,24 @@ resolve_info_uri (YelpUri *uri)
 
     if (!infopath) {
         /* Initialize infopath only once */
+
+        /* Use the same logic as the info program. If INFOPATH is not
+           specified, use the default. If it is specified, just use it
+           unless it ends with a colon, in which case we add the
+           default as a suffix.
+        */
         const gchar *env = g_getenv ("INFOPATH");
+        gchar *paths;
         if (!env || env[0] == '\0')
-            env = "/usr/info:/usr/share/info:/usr/local/info:/usr/local/share/info";
-        infopath = g_strsplit (env, ":", 0);
+            paths = g_strdup (default_info_path);
+        else if (env[strlen (env)-1] == ':')
+            paths = g_strconcat (env, default_info_path, NULL);
+        else
+            paths = g_strdup (env);
+
+        infopath = g_strsplit (paths, ":", 0);
+
+        g_free (paths);
     }
 
     colon = strchr (priv->res_arg, ':');
@@ -978,6 +1170,18 @@ resolve_xref_uri (YelpUri *uri)
         }
     }
 
+    if (priv->page_id &&
+        g_str_has_prefix (priv->docuri, "info:")) {
+        /*
+          Special characters get url-encoded when they get clicked on
+          as links. Info files, at least, don't want that so decode
+          the url again here.
+         */
+        gchar* tmp = priv->page_id;
+        priv->page_id = decode_url (priv->page_id);
+        g_free (tmp);
+    }
+
     if (g_str_has_prefix (priv->docuri, "ghelp:"))
         priv->fulluri = g_strconcat (priv->docuri,
                                      priv->page_id ? "?" : "",
@@ -1074,6 +1278,19 @@ resolve_gfile (YelpUri *uri, const gchar *hash)
                         priv->frag_id = g_strdup (splithash[1]);
                 }
             }
+            else if (yelp_settings_get_editor_mode (yelp_settings_get_default ())) {
+                g_object_unref (child);
+                child = g_file_get_child (priv->gfile, "index.page.stub");
+                if (g_file_query_exists (child, NULL)) {
+                    priv->tmptype = YELP_URI_DOCUMENT_TYPE_MALLARD;
+                    if (splithash) {
+                        if (priv->page_id == NULL)
+                            priv->page_id = g_strdup (splithash[0]);
+                        if (priv->frag_id == NULL && splithash[0])
+                            priv->frag_id = g_strdup (splithash[1]);
+                    }
+                }
+            }
             g_object_unref (child);
         }
         else {
@@ -1181,7 +1398,9 @@ resolve_gfile (YelpUri *uri, const gchar *hash)
     if (priv->fulluri == NULL)
         priv->fulluri = g_strconcat (priv->docuri,
                                      (priv->page_id || priv->frag_id) ? "#" : NULL,
-                                     priv->page_id ? priv->page_id : priv->frag_id,
+                                     priv->page_id ? priv->page_id : "",
+                                     priv->frag_id ? "#" : NULL,
+                                     priv->frag_id ? priv->frag_id : NULL,
                                      NULL);
 
     g_object_unref (info);
